/*
-----===== Arduino Web radio player ====----

(c)05 October 2014 by Vassilis Serasidis

Home: http://www.serasidis.gr
email: avrsite@yahoo.gr
Version: v1.0

Hardware:
- Any Arduino (I prefer Arduino pro mini or nano because of its small size and price).
- ENC28J60 ethernet module  (EtherCard library is used   - Written by Jean-Claude Wippler)
- VS1053B MP3 module/shield (VS1053B library is used     - Written by J. Coliz)
- Nokia 5110 LCD module     (PCD8544_mod library is used - Written by Carlos Rodrigues - Modified by Vassilis Serasidis )

Additional libraries:
- TimerOne (Software watchdog timer - Originaly written by Jesse Tane,
last modification Oct 2011 by Andrew Richards).

Description:
- This web radio player has 14 pre-defined web radio stations (station1,
station2,...,station13, station14).
You can add your favorite webradio stations. There is much free available
flash memory.

- For more web radio stations visit http://www.internet-radio.com/
Visit the http://www.internet-radio.com/ , choose the stations with up
to 64-kbps and copy the IP and the Port of the web radio.
For example, right-click on your favorite 64-kbps webradio icon and
select <save link location. Then, paste it to a text file.
You will see somethink like that:->
http://servers.internet-radio.com/tools/playlistgenerator/?u=http://108.163.215.90:8006/listen.pls&t=.pls
Replace the current station "station1_IP[] = {205,164,36,153}" with the
new station "station1_IP[] = {108,163,215,90}".
Do the same to the port. Replace the "station1_Port = 80" with the new
one "station1_Port = 8006".

You can replace the whole 14 stations with the stations of your choice.

** This circuit doesn't use a big RAM buffer. That causes a small lag (de-synchronize) due to delay transmission of data packets from the webradio server to your webradio player.

** I will try to improve the source code for supporting 128-kbps web
radio stations.

***** This source code and hardware are provided as is without any
warranty under GNU GPL v3 licence *****

     Have a good listening!
*/

#include <avr/pgmspace.h>
#include <EtherCard.h>
#define USE_VS1053 1
#include <VS1053.h>
#include <SPI.h>
#include <Wire.h>
#include <LiquidCrystal.h>
#include <LCDKeypad.h>
#include <TimerOne.h>
// #include <SdFat.h>
// #include <SdFatUtil.h>
#include <EEPROM.h>
#include "EEPROM_rw_anything.h"

#define DL(x) Serial.print(x)
#define DLn(x) Serial.println(x)
#define DLF(x) Serial.print(F(x))
#define DLFn(x) Serial.println(F(x))
#define DV(m, v) do{Serial.print(m);Serial.print(v);Serial.print(" ");}while(0)
#define DVn(m, v) do{Serial.print(m);Serial.println(v);}while(0)
#define DVF(m, v) do{Serial.print(F(m));Serial.print(v);Serial.print(F(" "));}while(0)
#define DVFn(m, v) do{Serial.print(F(m));Serial.println(v);}while(0)

#define BUFFER_LENGTH 600       //Ethernet data buffer length.
#define BUFFER_LENGTH2 32       //VS1053 data buffer length
#define LED1    A1

byte seconds;
boolean ViewStationInfo = false;
byte Ethernet::buffer[BUFFER_LENGTH];
byte MP3_buffer[BUFFER_LENGTH2];
int indexCounter;
static uint32_t timer;
#if VS_RELAY_NEEDED
static vs_reset_timer;
#endif

int8_t radioStationNumber; //Initial webradio station
boolean radioStationIsChanged;
boolean receivedData = true;

#define ETH_CS 3
static byte mymac[] = { 0x74,0x69,0x69,0x2D,0x30,0xF9 };

#define VS_CS A3
// VS1053  player(9, 6, 7, 8);  // cs_pin, dcs_pin, dreq_pin, reset_pin
// VS1053 xrst - Arduino D8 MOVE to A5
// VS1053 xdcs - Arduino D7 MOVE to A4
// VS1053 xcs  - Arduino D6 MOVE to A3
// VS1053 dreq - Arduino D2
VS1053  player(VS_CS, A4, 2, A5);  // cs_pin, dcs_pin, dreq_pin, reset_pin
// this VS_* stuff is any ugly hack to "fix" some mystery Ebay-sourced
// VS1053 boards.  I discovered that mine would only make sound if I
// lifted the ground for about 2 seconds *after* the audio stream had
// started going to the VS1053.  So I hacked in a relay and a 2 second
// timer to lift the VS1053 ground after the first (after a reset or power
// up) station starts playing.
//
// you can just ignore all of this.  Don't hook up the relay, and the code
// itself is harmless if you don't have this problem.
// relay drive is active low from pin VS_GND_RELAY
// relay NC contact goes to VS1053 ground, relay "common" goes to ground
// relay NO is no connection.

#if VS_RELAY_NEEDED
#define VS_GND_RELAY A2
byte VS_is_reset;
byte VS_timer_started;
#endif

LCDKeypad lcd;
// which does this:
// LCDKeypad::LCDKeypad() : LiquidCrystal(8, 9, 4, 5, 6, 7)

#if STATIONS_FILE
// this SdFat pig takes 321 bytes of RAM:
#define SD_SEL A2
// the object for the SD card file system:
SdFat sd;
#endif

typedef struct config_t {
    uint8_t last_radio_station;
} CONFIGBLK;

CONFIGBLK configuration;

typedef struct station {
    byte ip[4];
    unsigned int port;
} STATION;

const STATION stations[] PROGMEM = {
	{ { 10,  0,  0,  0},    80 },	//Dead Air
	{ {205,164, 36,153},    80 },	//BOM PSYTRANCE (1.FM TM)  64-kbps
	{ {205,164, 62, 15}, 10032 },	//1.FM - GAIA, 64-kbps
	{ {109,206, 96, 11},    80 },	//TOP FM Beograd 106,8  64-kpbs
	{ { 85, 17,121,216},  8468 },	//RADIO LEHOVO 971 GREECE, 64-kbps
	{ { 85, 17,121,103},  8800 },	//STAR FM 88.8 Corfu Greece, 64-kbps
	// 6
	{ { 85, 17,122, 39},  8530 },	//www.stylfm.gr laiko, 64-kbps
	{ {144, 76,204,149},  9940 },	// RADIO KARDOYLA - 64-kbps 22050 Hz
	{ {198, 50,101,130},  8245 },	//La Hit Radio, Rock - Metal - Hard Rock, 32-kbps
	{ { 94, 23, 66,155},  8106 },	// *ILR CHILL & GROOVE* 64-kbps
/*10*/	{ {205,164, 62, 22},  7012 },	//1.FM - ABSOLUTE TRANCE (EURO) RADIO   64-kbps

	{ {205,164, 62, 13}, 10144 },	//1.FM - Sax4Ever   64-kbps
	{ { 83,170,104, 91}, 31265 },	//Paradise Radio 106   64-kbps
	{ {205,164, 62, 13}, 10152 },	//Costa Del Mar - Chillout (1.FM), 64-kbps
	{ { 46, 28, 48,140},  9998 },	//AutoDJ, latin, cumbia, salsa, merengue, regueton, pasillos , 48-kbps
/*15*/	{ { 50,  7,173,162},  8116 },	// Big B Radio #CPOP - Asian Music - 64k

	{ { 50,  7,173,162},  8097 },	// Big B Radio #AsianPop - 64kbps
	{ {195,154,167, 62},  7264 },	// Radio China - 48kbps
	{ {198,154,106,104},  8985 },	//radioICAST.com Acid Jazz Blues Rock 96K
	{ { 69, 64, 33,125},  8016 },	//Jakarta Radio Cakrawala Mandarin 128k
/*20*/	{ {199,180, 73,130},  9070 },	//GFM OMG Chinese & Mandarin Christian Pop 128k
	{ {192,241,214,243},  8394 },	//DRHK Dig Rad Hong Kong - 96kbps
	{ {198,178,123, 11},  7292 } 	//Hong Kong Dubai Phillipines AutoDJ 64kbps
    };

#define LAST_STATION_NUMBER ((sizeof(stations)/sizeof(STATION)) - 1)
uint8_t last_station_number = LAST_STATION_NUMBER;

#define add_one_mms_station 0
#if add_one_mms_station
// One more station to test mms: streaming
// mms://live.ahradio.com.cn/xwgb
char test_url[] PROGMEM = "live.ahradio.com.cn";
char test_path[] PROGMEM = "/xwgb";
// but mms: is deprecated, now is "rtsp" protocol on 554
int  test_port = 1755;
// int  test_port = 554;

#endif

// mms: streams aren't going to work without a lot of work, and
// may never work on something as low powered as an Arduino.

//============================================================================
// called when the client request is complete
//============================================================================
static void my_callback (byte status, word off, word len) {
unsigned int i;
char *cp1, *cp2, lbuf[33];
int sp;
byte lcd_title_found, j;

    lcd_title_found = 0;
    if((indexCounter < 500) && (ViewStationInfo == false)) {
       if((cp1 = strstr_P((const char *)&Ethernet::buffer[off], PSTR("icy-name"))) != 0) {
	    cp1 += 9;
	    strncpy(lbuf, cp1, 32);
	    if((cp2 = strstr_P((const char *)lbuf, PSTR("\r"))) != 0) {
		*cp2 = '\0';
		j = strlen(lbuf);
		cp1 += j + 12;
		strncpy(cp2, cp1, 32 - j);
	    }
	    if((cp2 = strstr_P((const char *)lbuf, PSTR("\r"))) != 0) {
		*cp2 = '\0';
	    }
	    lcd.clear();
	    lcd.setCursor(0, 0); //LCD line 0
	    for(j = 0 ; j < 16 && lbuf[j] ; j++) {
		lcd.write(lbuf[j]);
	    }
	    lcd.setCursor(0, 1); //LCD line 1
	    for( ; j < 32 && lbuf[j] ; j++) {
		lcd.write(lbuf[j]);
	    }
	    lcd_title_found = 1;
       }
       for(i=0;i<len;i++) {
	   Serial.write(Ethernet::buffer[off+i]); //Show the web radio channel information (name, genre, bit rate etc).
	   if(!lcd_title_found) {
	       lcd.clear();
	       lcd.setCursor(0, 0); //LCD line 0
	       if((indexCounter + i > 179)
	       && (indexCounter + i < 301)
	       && Ethernet::buffer[off+i] >= ' ') {
		   lcd.write(Ethernet::buffer[off+i]);
	       }
	   }
	
	   if(Ethernet::buffer[off+i] == 0x0d &&
	      Ethernet::buffer[off+i+1] == 0x0a &&
	      Ethernet::buffer[off+i+2] == 0x0d &&
	      Ethernet::buffer[off+i+3] == 0x0a)
	      {
		  ViewStationInfo = true;
		  // Serial.println(F("    ViewStationInfo true"));
		  break; //We found the index end (0x0d,0x0a,0x0d,0x0a). Do not search anymore.
	   }
	}
	indexCounter += len;
    }
    else {
       uint8_t* data = (uint8_t *) Ethernet::buffer + off; //Get the data stream from ENC28J60 and...
       player.playChunk(data, len);	//...send them to VS1053B
       timer = millis();		//Update the timeout timer.
       receivedData = true;
    }
}

#if STATIONS_FILE
//===========================================================================
void parse_record(char *lbuf) {
STATION st;
char i, state;

    state = 0;
    memset((void *)&st, '\0', sizeof(STATION));
    for(i = 0 ; lbuf[i] ; i++) {
	if(!lbuf[i]) { break; }
	else if(lbuf[i] == ',') { state++; continue; }
	else if(isdigit(lbuf[i])) {
	    if(state < 4) {
		st.ip[state] *= 10;
		st.ip[state] += lbuf[i] - '0';
	    }
	    else {
		st.port *= 10;
		st.port += lbuf[i] - '0';
	    }
	}
    }
    lbuf[0] = '\0';
    // Serial.println(F("station is:"));
    for(i = 0 ; i < 4 ; i++) {
	Serial.print(st.ip[i], DEC);
	Serial.print((i == 3) ? F(":") : F(","));
    }
    Serial.println(st.port, DEC);
    return;
}

//===========================================================================
void read_stations_file (void) {
char fileName[13];
int c;
char lp, skip_to_nl;
char lbuf[40];
    skip_to_nl = 0;
    lp = 0;
    strcpy_P(fileName, PSTR("stations.txt"));
#if READ_STATIONS_TEST
    // for testing parse_record:
    strcpy_P(lbuf, PSTR(" 205,164, 36,153,    80 # BOM PSYTRANCE"));
#else
    ifstream file(fileName);
    if (!file.is_open()) Serial.println(F("error opening file"));
    while((c = file.get()) >= 0) {
	if(c == '#') { // comment
	    skip_to_nl = 1;
	    continue;
	}
	else if(c == '\r' || c == '\n') { // new line
	    skip_to_nl = 0;
	    lbuf[lp] = '\0';
	    if(lbuf[0]) { parse_record(lbuf); }
	    continue;
	}
	else if(skip_to_nl == 1) {
	    continue;
	}
	else {
	    lbuf[lp] = c;
	    if(lp < 40) { lp++; }
	}
    }
#endif
    if(lbuf[0]) { parse_record(lbuf); }
}
#endif

//===========================================================================
void setup() {
    pinMode(LED1, OUTPUT);	//LED1 is connected to A1 pin

    Serial.begin(57600);	//Start serial port with 57600 bits per seconds
    pinMode(VS_CS, OUTPUT);
    digitalWrite(VS_CS, HIGH);	// turn on pullup
    pinMode(ETH_CS, OUTPUT);
    digitalWrite(ETH_CS, HIGH);	// turn on pullup
    SPI.begin();		//Start Serial Peripheral Interface (SPI)
#if STATIONS_FILE
    //Initialize the SdCard.
    pinMode(SD_SEL, OUTPUT);
    digitalWrite(SD_SEL, HIGH);	// turn on pullup
    if(!sd.begin(SD_SEL, SPI_FULL_SPEED)) sd.initErrorHalt();
    // depending upon your SdCard environment, SPI_HALF_SPEED may work better.
    if(!sd.chdir("/")) sd.errorHalt("sd.chdir");
    read_stations_file();
#endif

#if VS_RELAY_NEEDED
    pinMode(VS_GND_RELAY, OUTPUT);	// relay to lift ground on "funny" VS1053
    digitalWrite(VS_GND_RELAY, HIGH);	// ground the VS1053
#endif
    indexCounter = 0;
    lcd.begin(16, 2); //2 rows x 16 chars text
    lcd.clear();
    lcd.setCursor(0, 0);	// column, row
    lcd.print(F("WEBRADIO PLAYER"));

    Serial.println(F("\n--== Arduino WebRadio player ==--\n"));
    Serial.println(F("(c)2014 by Vassilis Serasidis"));
    Serial.println(F("http://www.serasidis.gr\n"));
    Serial.println(F("Initializing VS1053..."));
    lcd.setCursor(0, 1);	// column, row
    lcd.print(F("Init VS1053..."));

    // delay(2000);
    player.begin();		//Start VS1053B
    player.setVolume(0);    //Set the volume to the maximum.

    Serial.println(F("Connecting..."));
    lcd.setCursor(0, 1);	// column, row
    lcd.print(F("Connecting..."));
    //Initiallize ENC28J60. Chip Select pin (CS) is on arduino pin D10.
    while (ether.begin(sizeof Ethernet::buffer, mymac, ETH_CS) == 0) {
	Serial.println(F("Failed to access Ethernet controller"));
	lcd.setCursor(0, 1);	// column, row
	lcd.print(F("ether failed"));
	delay(10);
    }
    lcd.setCursor(0, 1);	// column, row
    lcd.print(F("ethernet OK "));
    Serial.println(F("got past ether.begin()"));
    while (!ether.dhcpSetup()) {
	Serial.println(F("DHCP failed"));
	lcd.setCursor(0, 1);	// column, row
	lcd.print(F("DHCP failed"));
	delay(10);
    }
    Serial.println(F("got past dhcpSetup()"));
    lcd.setCursor(0, 1);	// column, row
    lcd.print(F("DHCP ok     "));

    ether.persistTcpConnection(true); //Keep TCP/IP connection alive.
    // ViewStationInfo = true;

#if VS_RELAY_NEEDED
    VS_is_reset = 0;
    VS_timer_started = 0;
#endif
    seconds = 0;
    // set a timer of length 5000000 microseconds (or 5 sec => the
    // LED will blink 5 times, 5 cycles of on-and-off, per second)
    // changed to one second ISR, which runs the "lost signal" code every 5 secs
    Timer1.initialize(1000000);
    Timer1.attachInterrupt( timerIsr ); // attach the service routine here
    EEPROM_readAnything(0, configuration);
    radioStationIsChanged = true;
    radioStationNumber = configuration.last_radio_station ;
    if(radioStationNumber > last_station_number
    || radioStationNumber < 1) {
	radioStationNumber = 1;
    }
#if add_one_mms_station
    last_station_number++;
    radioStationNumber = last_station_number - 1;
#endif
}

//=============================================================================
//
//=============================================================================
void loop() {
int buttonPressed;
char pnum[4];
STATION stemp;

#if VS_RELAY_NEEDED
    if(!VS_is_reset && (ViewStationInfo == true) && !VS_timer_started) {
	digitalWrite(VS_GND_RELAY, LOW);	// un-ground the "funny" VS1053
	vs_reset_timer = millis();
	VS_timer_started = 1;
    }
#endif
    ether.packetLoop(ether.packetReceive()); //Read ENC28J60 for new incoming data packet.

   buttonPressed=lcd.button();
   //Check if the 'next' switch has been pressed.
   if(buttonPressed == KEYPAD_RIGHT) {
     while(buttonPressed == KEYPAD_RIGHT) {
       delay(100);
       buttonPressed=lcd.button();
     }
     radioStationNumber++;
     goto Increment;
   }

   if(buttonPressed == KEYPAD_UP) {
     while(buttonPressed == KEYPAD_UP) {
       delay(100);
       buttonPressed=lcd.button();
     }
     radioStationNumber += 5;
Increment:
     radioStationIsChanged = true;
     if(radioStationNumber > last_station_number) {
       radioStationNumber -= last_station_number ;
     }
   }
   //Check if the 'previous' switch has been pressed.
   if(buttonPressed== KEYPAD_LEFT) {
       while(buttonPressed == KEYPAD_LEFT) {
         delay(100);
         buttonPressed=lcd.button();
       }
       radioStationNumber -= 1;
       goto Decrement;
   }
   if(buttonPressed == KEYPAD_DOWN) {
     while(buttonPressed == KEYPAD_DOWN) {
       delay(100);
       buttonPressed=lcd.button();
     }
     radioStationNumber -= 5;
Decrement:
     radioStationIsChanged = true;
     if(radioStationNumber < 1) {
       radioStationNumber += last_station_number;
     }
   }
   if(buttonPressed == KEYPAD_SELECT) {	// STOP
      radioStationNumber = 0;
     radioStationIsChanged = true;
   }

   //If 'next' or 'previous' switch has been pressed, play the selected webradio station.
    if(radioStationIsChanged == true) {

	configuration.last_radio_station = radioStationNumber;
	EEPROM_writeAnything(0, configuration);

	sprintf(pnum, "%2d", radioStationNumber);

#if add_one_mms_station
	if(radioStationNumber > LAST_STATION_NUMBER) {
	    playWebRadioStationURL(test_url, test_path, test_port, pnum);
	}
	else {
#endif
	    // copy STATION data from PROGMEM to stemp in ram:
	    for (unsigned char k = 0; k < 4; k++) {
		stemp.ip[k] = pgm_read_byte(&(stations[radioStationNumber].ip[k]));
	    }
	    stemp.port = pgm_read_word(&(stations[radioStationNumber].port));

	    playWebRadioStation(stemp.ip, stemp.port, pnum);
#if add_one_mms_station
	}
#endif
    }

    radioStationIsChanged = false;

    // Timeout.  If the song stops playing for 5 seconds, re-connect.
    if((millis() > timer + 5000)) {
	radioStationIsChanged = true;
	ViewStationInfo = false;
	Serial.print(F("\nre-connecting to the server...\n"));
	lcd.clear();
	lcd.setCursor(0,0);
	lcd.print(F("re-connecting to"));
	lcd.setCursor(0,1);
	lcd.print(F("the server"));
	timer = millis();
    }
}

//============================================================================
//
//============================================================================
void playWebRadioStation ( byte ip[4], const int hisPort, char* preset ) {
    Serial.print(F("\n\n<")); //Print the station info to the serial port
    Serial.print(preset);
    Serial.println(F("> ============================================================"));

    player.stopSong();
    ViewStationInfo = false;
    indexCounter = 0;
    ether.copyIp(ether.hisip, ip);
    ether.hisport = hisPort;
    ether.printIp(F("IP:   "), ether.myip);
    ether.printIp(F("GW:   "), ether.gwip);
    ether.printIp(F("DNS:  "), ether.dnsip);
    ether.printIp(F("SRV:  "), ether.hisip);
    Serial.print(F("Port: "));
    Serial.println(ether.hisport);
    Serial.println();
    //ether.browseUrl(PSTR("/"), "",PSTR(""), PSTR("Icy-MetaData:1"), my_callback);
    ether.browseUrl(PSTR("/"), "",PSTR(""), PSTR(""), my_callback);

    lcd.setCursor(0, 0); //LCD line 0
    lcd.write('<');
    lcd.print(preset); //Show on LCD the webradio number (1-14)
    lcd.print(F(">            "));
}

//============================================================================
#if add_one_mms_station
void playWebRadioStationURL ( const char *url, const char *path, const int hisPort, char* preset ) {
char ubuf[30];
    Serial.print(F("\n\n<")); //Print the station info to the serial port
    Serial.print(preset);
    Serial.println(F("> ============================================================"));

    player.stopSong();
    ViewStationInfo = false;
    indexCounter = 0;
    strncpy_P(ubuf, url, 30);
    strncat_P(ubuf, path, 30 - strlen(ubuf));
    DVFn("pWRSURL = ", ubuf);
    if(!ether.dnsLookup(url, false)) {	// if successful, sets ether.hisip
	Serial.println(F("DNS failed. Unable to continue."));
	while (true);		// die!
    }
    ether.hisport = hisPort;
    ether.printIp(F("IP:   "), ether.myip);
    ether.printIp(F("GW:   "), ether.gwip);
    ether.printIp(F("DNS:  "), ether.dnsip);
    ether.printIp(F("SRV:  "), ether.hisip);
    Serial.print(F("Port: "));
    Serial.println(ether.hisport);
    Serial.println();
// ether.browseUrl(PSTR("/weather/ASXX0075:1"), "", website, browseUrlCallback);
    ether.browseUrl(path, PSTR(""), url, my_callback);

    lcd.setCursor(0, 0); //LCD line 0
    lcd.write('<');
    lcd.print(preset); //Show on LCD the webradio number (1-14)
    lcd.print(F(">            "));
}
#endif

#if VS_RELAY_NEEDED
void one_second_isr (void) {
    if(VS_timer_started) {
	if((millis() > vs_reset_timer + 2000)) {
	    digitalWrite(VS_GND_RELAY, HIGH);	// ground the VS1053
	    VS_is_reset = 1;
	    VS_timer_started = 0;
	}
    }
}
#endif
//============================================================================
// Custom ISR Timer Routine
//============================================================================
void timerIsr() {
    // Toggle LED
    seconds++;
    if(seconds >= 5) {
	seconds = 0;
	digitalWrite( LED1, digitalRead( LED1 ) ^ 1 );
	if(receivedData == true) {
	    receivedData = false;
	}
	else {
	    radioStationIsChanged = true;
	    //Initiallize ENC28J60. Chip Select pin (CS) is connected on arduino pin D10.
	    if (ether.begin(sizeof Ethernet::buffer, mymac, ETH_CS) == 0) {
		Serial.println(F("Failed to access Ethernet controller"));
		lcd.setCursor(0, 1);	// column, row
		lcd.print(F("ethernet failed"));
	    }
	    else {
		lcd.setCursor(0, 1);	// column, row
		lcd.print(F("ethernet OK "));
	    }
	    if (!ether.dhcpSetup()) {
		Serial.println(F("DHCP failed"));
		lcd.setCursor(0, 1);	// column, row
		lcd.print(F("DHCP failed"));
	    }
	    ether.persistTcpConnection(true); //Keep TCP/IP connection alive.
	}
    }
#if VS_RELAY_NEEDED
    one_second_isr();
#endif
}

